"""Django admin functionality that is relevant to the entire app"""
from django.contrib import admin


class TimestampedModelAdmin(admin.ModelAdmin):
    """
    A ModelAdmin that includes timestamp fields in the detail view and, optionally, in the list view
    """

    include_timestamps_in_list = False
    include_created_on_in_list = False

    @staticmethod
    def _join_and_dedupe(existing_field_names, field_names_to_add):  # pragma: no cover
        """
        Joins two tuples of field names together, and ensures that no duplicate field names are added

        Args:
            existing_field_names (Tuple[str]): Field names
            field_names_to_add (Tuple[str]): Field names to add to the existing ones

        Returns:
            Tuple[str]: The combined field names without any duplicates, unless there were any duplicates in the
                tuple of existing field names
        """
        return existing_field_names + tuple(
            field for field in field_names_to_add if field not in existing_field_names
        )

    def get_list_display(self, request):  # pragma: no cover
        list_display = tuple(super().get_list_display(request) or ())
        added_fields = ()
        if self.include_timestamps_in_list:
            added_fields += ("created_on", "updated_on")
        elif self.include_created_on_in_list:
            added_fields += ("created_on",)
        return self._join_and_dedupe(list_display, added_fields)

    def get_readonly_fields(self, request, obj=None):  # pragma: no cover
        readonly_fields = tuple(super().get_readonly_fields(request, obj=obj) or ())
        if obj is None:
            return readonly_fields
        return self._join_and_dedupe(readonly_fields, ("created_on", "updated_on"))

    def get_exclude(self, request, obj=None):  # pragma: no cover
        exclude = tuple(super().get_exclude(request, obj=obj) or ())
        return self._join_and_dedupe(exclude, ("created_on", "updated_on"))



"""Templatetags for rendering webpack bundle script tags"""
from os import path
from typing import Any, Dict, Iterator

from django import template
from django.conf import settings
from django.http import HttpRequest
from django.utils.safestring import SafeText, mark_safe
from webpack_loader.utils import get_loader

from mitol.common.utils.webpack import webpack_public_path

register = template.Library()


def _get_bundle(request: HttpRequest, bundle_name: str) -> Iterator[dict]:
    """
    Update bundle URLs to handle webpack hot reloading correctly if DEBUG=True

    Args:
        request (django.http.request.HttpRequest): A request
        bundle_name (str): The name of the webpack bundle

    Returns:
        iterable of dict:
            The chunks of the bundle. Usually there's only one but I suppose you could have
            CSS and JS chunks for one bundle for example
    """
    if not settings.WEBPACK_DISABLE_LOADER_STATS:
        for chunk in get_loader("DEFAULT").get_bundle(bundle_name):
            chunk_copy = dict(chunk)
            chunk_copy["url"] = path.join(webpack_public_path(request), chunk["name"])
            yield chunk_copy


@register.simple_tag(takes_context=True)
def render_bundle(
    context: Dict[str, Any], bundle_name: str, added_attrs: str = ""
) -> SafeText:
    """
    Render the script tags for a Webpack bundle

    We use this instead of webpack_loader.templatetags.webpack_loader.render_bundle because we want to substitute
    a dynamic URL for webpack dev environments. Maybe in the future we should refactor to use publicPath
    instead for this.

    Args:
        context (dict): The context for rendering the template (includes request)
        bundle_name (str): The name of the bundle to render
        added_attrs (str): Optional string of HTML attributes to add to the script/link tag

    Returns:
        django.utils.safestring.SafeText:
    """
    try:
        bundle = _get_bundle(context["request"], bundle_name)
        return _render_tags(bundle, added_attrs)
    except OSError:
        # webpack-stats.json doesn't exist
        return mark_safe("")


def _render_tags(bundle: Iterator[dict], added_attrs: str = "") -> SafeText:
    """
    Outputs tags for template rendering.
    Adapted from webpack_loader.utils.get_as_tags and webpack_loader.templatetags.webpack_loader.

    Args:
        bundle (iterable of dict): The information about a webpack bundle
        added_attrs (str): Optional string of HTML attributes to add to the script/link tag

    Returns:
        django.utils.safestring.SafeText: HTML for rendering bundles
    """

    tags = []
    for chunk in bundle:
        if chunk["name"].endswith((".js", ".js.gz")):
            tags.append(
                ('<script type="text/javascript" src="{}" {} ></script>').format(
                    chunk["url"], added_attrs
                )
            )
        elif chunk["name"].endswith((".css", ".css.gz")):
            tags.append(
                ('<link type="text/css" href="{}" rel="stylesheet" {} />').format(
                    chunk["url"], added_attrs
                )
            )
    return mark_safe("\n".join(tags))



"""Common app AppConfigs"""
import os
from typing import List

from django.apps import AppConfig
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured


class BaseApp(AppConfig):
    """Base application class"""

    required_settings: List[str] = []

    def validate_required_settings(self):
        """
        Verify that the required settings have been set.

        This allows verifying settings that an application requires, but
        aren't necessarily covered by a call to `mitol.common.envs` functions.
        """
        missing_settings = []

        for setting_name in self.required_settings:
            if not hasattr(settings, setting_name):
                missing_settings.append(setting_name)

        if missing_settings:
            raise ImproperlyConfigured(
                "The following settings are missing: {}. You need to add these environment variables in .env file.".format(
                    ", ".join(missing_settings)
                )
            )

    def ready(self):
        """The application is ready"""
        self.validate_required_settings()


class CommonApp(AppConfig):
    """Default configuration for the common app"""

    name = "mitol.common"
    label = "common"
    verbose_name = "Common"

    # necessary because this is a namespaced app
    path = os.path.dirname(os.path.abspath(__file__))



"""Currency utilities"""
from decimal import Decimal


def format_price(amount: Decimal) -> str:
    """
    Format a price in USD

    Args:
        amount (decimal.Decimal): A decimal value

    Returns:
        str: A currency string
    """
    return f"${amount:0,.2f}"



"""Common settings"""
from mitol.common.envs import app_namespaced, get_string

SITE_BASE_URL = get_string(
    name=app_namespaced("BASE_URL"),
    default=None,
    description="Base url for the application in the format PROTOCOL://HOSTNAME[:PORT]",
    required=True,
)



"""Common factories"""

from django.conf import settings
from django.utils.module_loading import import_string

UserFactory = import_string(
    getattr(
        settings,
        "MITOL_COMMON_USER_FACTORY",
        "mitol.common.factories.defaults.UserFactory",
    )
)



from functools import wraps
import random

from django.utils.cache import get_max_age, patch_cache_control


def cache_control_max_age_jitter(*args, **kwargs):
    def _cache_controller(viewfunc):
        @wraps(viewfunc)
        def _cache_controlled(request, *args, **kwargs):
            # Ensure argument looks like a request.
            if not hasattr(request, "META"):
                raise TypeError(
                    "cache_control_max_age_jitter didn't receive an HttpRequest. If you are "
                    "decorating a classmethod, be sure to use "
                     "@method_decorator."
                )
            response = viewfunc(request, *args, **kwargs)
            max_age = get_max_age(response)
            # add random delay upto 5 minutes
            if max_age:
                kwargs["max_age"] = max_age + random.randint(1, 600)
                patch_cache_control(response, **kwargs)
            return response
        return _cache_controlled
    return _cache_controller



"""Collections utilities"""
from collections.abc import Iterable
from itertools import groupby, islice, tee


def dict_without_keys(d, *omitkeys):
    """
    Returns a copy of a dict without the specified keys

    Args:
        d (dict): A dict that to omit keys from
        *omitkeys: Variable length list of keys to omit

    Returns:
        dict: A dict with omitted keys
    """
    return {key: d[key] for key in d.keys() if key not in omitkeys}


def filter_dict_by_key_set(dict_to_filter, key_set):
    """Takes a dictionary and returns a copy with only the keys that exist in the given set"""
    return {key: dict_to_filter[key] for key in dict_to_filter.keys() if key in key_set}


def first_matching_item(iterable, predicate):
    """
    Gets the first item in an iterable that matches a predicate (or None if nothing matches)

    Returns:
        Matching item or None
    """
    return next(filter(predicate, iterable), None)


def find_object_with_matching_attr(iterable, attr_name, value):
    """
    Finds the first item in an iterable that has an attribute with the given name and value. Returns
    None otherwise.

    Returns:
        Matching item or None
    """
    for item in iterable:
        try:
            if getattr(item, attr_name) == value:
                return item
        except AttributeError:
            pass
    return None


def has_equal_properties(obj, property_dict):
    """
    Returns True if the given object has the properties indicated by the keys of the given dict, and the values
    of those properties match the values of the dict
    """
    for field, value in property_dict.items():
        try:
            if getattr(obj, field) != value:
                return False
        except AttributeError:
            return False
    return True


def first_or_none(iterable):
    """
    Returns the first item in an iterable, or None if the iterable is empty

    Args:
        iterable (iterable): Some iterable
    Returns:
        first item or None
    """
    return next((x for x in iterable), None)


def max_or_none(iterable):
    """
    Returns the max of some iterable, or None if the iterable has no items

    Args:
        iterable (iterable): Some iterable
    Returns:
        max item or None
    """
    try:
        return max(iterable)
    except ValueError:
        return None


def partition(items, predicate=bool):
    """
    Partitions an iterable into two different iterables - the first does not match the given condition, and the second
    does match the given condition.

    Args:
        items (iterable): An iterable of items to partition
        predicate (function): A function that takes each item and returns True or False
    Returns:
        tuple of iterables: An iterable of non-matching items, paired with an iterable of matching items
    """
    a, b = tee((predicate(item), item) for item in items)
    return (item for pred, item in a if not pred), (item for pred, item in b if pred)


def partition_to_lists(items, predicate=bool):
    """
    Partitions an iterable into two different lists - the first does not match the given condition, and the second
    does match the given condition.

    Args:
        items (iterable): An iterable of items to partition
        predicate (function): A function that takes each item and returns True or False
    Returns:
        tuple of lists: A list of non-matching items, paired with a list of matching items
    """
    a, b = partition(items, predicate=predicate)
    return list(a), list(b)


def unique(iterable):
    """
    Returns a generator containing all unique items in an iterable

    Args:
        iterable (iterable): An iterable of any hashable items
    Returns:
        generator: Unique items in the given iterable
    """
    seen = set()
    return (x for x in iterable if x not in seen and not seen.add(x))


def unique_ignore_case(strings):
    """
    Returns a generator containing all unique strings (coerced to lowercase) in a given iterable

    Args:
        strings (iterable of str): An iterable of strings
    Returns:
        generator: Unique lowercase strings in the given iterable
    """
    seen = set()
    return (s for s in map(str.lower, strings) if s not in seen and not seen.add(s))


def item_at_index_or_none(indexable, index):
    """
    Returns the item at a certain index, or None if that index doesn't exist

    Args:
        indexable (list or tuple):
        index (int): The index in the list or tuple

    Returns:
        The item at the given index, or None
    """
    try:
        return indexable[index]
    except IndexError:
        return None


def item_at_index_or_blank(indexable, index):
    """
    Returns the item at a certain index, or a blank string if that index doesn't exist

    Args:
        indexable (List[str]): A list of strings
        index (int): The index in the list or tuple

    Returns:
        str: The item at the given index, or a blank string
    """
    return item_at_index_or_none(indexable, index) or ""


def all_equal(*args):
    """
    Returns True if all of the provided args are equal to each other

    Args:
        *args (hashable): Arguments of any hashable type

    Returns:
        bool: True if all of the provided args are equal, or if the args are empty
    """
    return len(set(args)) <= 1


def all_unique(iterable):
    """
    Returns True if all of the provided args are equal to each other

    Args:
        iterable: An iterable of hashable items

    Returns:
        bool: True if all of the provided args are equal
    """
    return len(set(iterable)) == len(iterable)


def has_all_keys(dict_to_scan, keys):
    """
    Returns True if the given dict has all of the given keys

    Args:
        dict_to_scan (dict):
        keys (iterable of str): Iterable of keys to check for

    Returns:
        bool: True if the given dict has all of the given keys
    """
    return all(key in dict_to_scan for key in keys)


def group_into_dict(items, key_fn):
    """
    Groups items into a dictionary based on a key generated by a given function

    Examples:
        items = [
            Car(make="Honda", model="Civic"),
            Car(make="Honda", model="Accord"),
            Car(make="Ford", model="F150"),
            Car(make="Ford", model="Focus"),
        ]
        group_into_dict(items, lambda car: car.make) == {
            "Honda": [Car(make="Honda", model="Civic"), Car(make="Honda", model="Accord")],
            "Ford": [Car(make="Ford", model="F150"), Car(make="Ford", model="Focus")],
        }

    Args:
        items (Iterable[T]): An iterable of objects to group into a dictionary
        key_fn (Callable[[T], Any]): A function that will take an individual item and produce a dict key

    Returns:
        Dict[Any, T]: A dictionary with keys produced by the key function paired with a list of all the given
            items that produced that key.
    """
    sorted_items = sorted(items, key=key_fn)
    return {
        key: list(values_iter) for key, values_iter in groupby(sorted_items, key=key_fn)
    }


def chunks(iterable: Iterable, *, chunk_size: int = 20):
    """
    Yields chunks of an iterable as sub lists each of max size chunk_size.

    Args:
        iterable (iterable): iterable of elements to chunk
        chunk_size (int): Max size of each sublist

    Yields:
        list: List containing a slice of list_to_chunk
    """
    chunk_size = max(1, chunk_size)
    iterable = iter(iterable)
    chunk = list(islice(iterable, chunk_size))

    while len(chunk) > 0:
        yield chunk
        chunk = list(islice(iterable, chunk_size))



"""Common default factories"""

from django.contrib.auth import get_user_model
from factory import Faker, Sequence
from factory.django import DjangoModelFactory


class UserFactory(DjangoModelFactory):
    """
    User factory for default django model

    This expects the user model to be django.contrib.auth.models.User or a subclass of it
    """

    username = Sequence(lambda n: f"user-{n}")
    email = Sequence(lambda n: f"user-{n}@example.com")
    first_name = Faker("first_name")
    last_name = Faker("first_name")

    class Meta:
        model = get_user_model()



"""Generates app.json based on settings configuration"""
import json

from django.core.management.base import BaseCommand

from mitol.common import envs


class Command(BaseCommand):
    """Generates app.json based on settings configuration"""

    help = "Generates app.json based on settings configuration"

    def handle(self, *args, **options):
        """Generates app.json based on settings configuration"""
        config = envs.generate_app_json()

        with open("app.json", "w") as app_json:
            app_json.write(json.dumps(config, sort_keys=True, indent=2))
        self.stdout.write(self.style.SUCCESS("Updated app.json"))



"""Pytest testing utils"""
import abc
import json
import traceback
from contextlib import contextmanager

import pytest

from mitol.common.envs import generate_app_json


def any_instance_of(*cls):
    """
    Returns a type that evaluates __eq__ in isinstance terms

    Args:
        cls (list of types): variable list of types to ensure equality against

    Returns:
        AnyInstanceOf: dynamic class type with the desired equality
    """

    class AnyInstanceOf(metaclass=abc.ABCMeta):
        """Dynamic class type for __eq__ in terms of isinstance"""

        def __eq__(self, other):
            return isinstance(other, cls)

    for c in cls:
        AnyInstanceOf.register(c)
    return AnyInstanceOf()


@contextmanager
def assert_not_raises():
    """Used to assert that the context does not raise an exception"""
    try:
        yield
    except AssertionError:
        raise
    except Exception:
        pytest.fail(f"An exception was not raised: {traceback.format_exc()}")


class MockResponse:
    """
    Mock requests.Response
    """

    def __init__(
        self, content, status_code=200, content_type="application/json", url=None
    ):
        if isinstance(content, (dict, list)):
            self.content = json.dumps(content)
        else:
            self.content = str(content)
        self.text = self.content
        self.status_code = status_code
        self.headers = {"Content-Type": content_type}
        if url:
            self.url = url

    def json(self):
        """Return json content"""
        return json.loads(self.content)


def test_app_json_modified():  # pragma: no cover
    """
    Pytest test that verifies app.json is up-to-date

    To use this, you should import this into a test file somewhere in your project:

    from mitol.common.pytest_utils import test_app_json_modified
    """
    with open("app.json") as app_json_file:
        app_json = json.load(app_json_file)

    generated_app_json = generate_app_json()

    # pytest will print the difference
    assert json.dumps(app_json, sort_keys=True, indent=2) == json.dumps(
        generated_app_json, sort_keys=True, indent=2
    ), "Generated app.json does not match the app.json file. Please use the 'generate_app_json' management command to update app.json"



"""
Common model classes
"""
import copy
from typing import Dict, Iterable, List, Type, TypeVar, Union

from django.db.models import DateTimeField, Model, prefetch_related_objects
from django.db.models.query import QuerySet

from mitol.common.utils.datetime import now_in_utc


class TimestampedModelQuerySet(QuerySet):
    """
    Subclassed QuerySet for TimestampedModel
    """

    def update(self, **kwargs):
        """
        Automatically update updated_on timestamp when .update(). This is because .update()
        does not go through .save(), thus will not auto_now, because it happens on the
        database level without loading objects into memory.
        """
        if "updated_on" not in kwargs:
            kwargs["updated_on"] = now_in_utc()
        return super().update(**kwargs)


class TimestampedModel(Model):
    """Base model for created_on/updated_on timestamp fields"""

    objects = TimestampedModelQuerySet.as_manager()
    created_on = DateTimeField(auto_now_add=True)  # UTC
    updated_on = DateTimeField(auto_now=True)  # UTC

    class Meta:
        abstract = True


class SingletonModel(Model):
    """Model class for models representing tables that should only have a single record"""

    def save(
        self, force_insert=False, force_update=False, using=None, update_fields=None
    ):
        if force_insert and self._meta.model.objects.count() > 0:
            raise ValidationError(
                "Only one {} object should exist. Update the existing object instead "
                "of creating a new one.".format(self.__class__.__name__)
            )
        return super().save(
            force_insert=force_insert,
            force_update=force_update,
            using=using,
            update_fields=update_fields,
        )

    class Meta:
        abstract = True


_ModelClass = TypeVar("_ModelClass", bound=Model)
_PrefetchGenericQuerySet = TypeVar(
    "_PrefetchGenericQuerySet", bound="PrefetchGenericQuerySet"
)


def _items_for_class(
    content_type_field: str, items: Iterable[_ModelClass], model_cls: Type[_ModelClass]
) -> Iterable[_ModelClass]:
    """Returns a list of items that matches a class by content_type"""
    return [
        item
        for item in items
        if getattr(item, content_type_field).model_class() == model_cls
    ]


class PrefetchGenericQuerySet(QuerySet):
    """QuerySet supporting for prefetching over generic relationships"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._prefetch_generic_related_lookups = {}
        self._prefetch_generic_done = False

    def prefetch_generic_related(
        self,
        content_type_field: str,
        model_lookups: Dict[
            Union[List[Type[_ModelClass]], Type[_ModelClass]], List[str]
        ],
    ) -> _PrefetchGenericQuerySet:
        """
        Configure prefetch_related over generic relations

        Args:
            content_type_field(str): the field name for the ContentType
            model_lookups(dict of (list of class or class, list of str)):
                a mapping of model classes to lookups

        Returns:
            QuerySet: the new queryset with prefetching configured

        """
        qs = self._chain()  # type: ignore

        for model_classes, lookups in model_lookups.items():
            model_classes = (
                model_classes
                if isinstance(model_classes, Iterable)
                else [model_classes]
            )
            for model_cls in model_classes:
                key = (content_type_field, model_cls)

                qs._prefetch_generic_related_lookups[key] = [
                    *qs._prefetch_generic_related_lookups.get(key, []),
                    *lookups,
                ]

        return qs

    def _prefetch_generic_related_objects(self):
        """Prefetch related objects on a per-model basis"""
        for (
            (content_type_field, model_cls),
            lookups,
        ) in self._prefetch_generic_related_lookups.items():
            items = _items_for_class(content_type_field, self._result_cache, model_cls)
            prefetch_related_objects(items, *lookups)
        self._prefetch_generic_done = True

    def _fetch_all(self):
        """Called when a query is evaluated"""
        # first fetch non-generic data, this avoid N+1 issues on the generic items themselves
        super()._fetch_all()

        if self._prefetch_generic_related_lookups and not self._prefetch_generic_done:
            self._prefetch_generic_related_objects()

    def _clone(self):
        """Clone the queryset"""

        c = super()._clone()
        c._prefetch_generic_related_lookups = copy.deepcopy(
            self._prefetch_generic_related_lookups
        )
        return c



"""Functions reading and parsing environment variables"""
import importlib
import json
import os
from ast import literal_eval
from functools import wraps
from typing import Any, Callable, Dict, List, NamedTuple, Union

from django.core.exceptions import ImproperlyConfigured


class EnvironmentVariableParseException(ImproperlyConfigured):
    """Environment variable was not parsed correctly"""


class EnvVariable(NamedTuple):
    """Representation of an environment variable"""

    name: str
    default: Any
    description: str
    required: bool
    dev_only: bool
    value: Any
    write_app_json: bool


def var_parser(parser_func: Callable):
    """
    Decorator to create a var parser func

    Args:
        parser_func (callable):
            a function that takes one argument which will be the raw value and
            returns a parsed value or raises an error
    """

    @wraps(parser_func)
    def wrapper(
        self,
        *,
        name: str,
        description: str,
        default: Any = None,
        required: bool = False,
        dev_only: bool = False,
        write_app_json: bool = True,
    ) -> EnvVariable:
        """
        Get an environment variable

        Args:
            name (str): An environment variable name
            default (str): The default value to use if the environment variable doesn't exist.
            description (str): The description of how this variable is used
            required (bool): Whether this variable is required at runtime
            dev_only (bool): Whether this variable is only applicable in dev environments
            write_app_json (bool): Whether this variable is written to app.json

        Raises:
            ValueError:
                If the environment variable args are incorrect

        Returns:
            any:
                The raw environment variable value
        """
        configured_envs = self._configured_vars
        environ = self._env

        if name in configured_envs:
            raise ValueError(f"Environment variable '{name}' was used more than once")

        value = environ.get(name, default)
        # attempt to parse the value before we store it in configured_envs
        # this ensures that get_any works since we don't store the various parse attempts until one succeeds
        value = parser_func(name, value, default)

        configured_envs[name] = EnvVariable(
            name, default, description, required, dev_only, value, write_app_json
        )

        return value

    return wrapper


def parse_bool(name: str, value: str, default: bool) -> bool:
    """
    Attempts to parse a bool

    Arguments:
        value (str or bool):
            the value as either an unparsed string or a bool in case of a default value

    Raises:
        EnvironmentVariableParseException:
            raised if the value wasn't parsable

    Returns:
        bool:
            parsed value
    """

    if isinstance(value, bool):
        return value

    parsed_value = value.lower()
    if parsed_value == "true":
        return True
    elif parsed_value == "false":
        return False

    raise EnvironmentVariableParseException(
        "Expected value in {name}={value} to be a boolean".format(
            name=name, value=value
        )
    )


def parse_int(name: str, value: str, default: int) -> int:
    """
    Attempts to parse a int

    Arguments:
        value (str or int):
            the value as either an unparsed string or an int in case of a default value

    Raises:
        EnvironmentVariableParseException:
            raised if the value wasn't parsable

    Returns:
        int:
            parsed value
    """

    if isinstance(value, int) or (value is None and default is None):
        return value

    try:
        parsed_value = int(value)
    except ValueError as ex:
        raise EnvironmentVariableParseException(
            "Expected value in {name}={value} to be an int".format(
                name=name, value=value
            )
        ) from ex

    return parsed_value


def parse_str(name: str, value: str, default: str):
    """
    Parses a str (identity function)

    Arguments:
        value (str):
            the value as either a str

    Returns:
        str:
            parsed value
    """
    return value


def parse_list_literal(
    name: str,
    value: Union[str, List[Any]],
    default: List[Any],
) -> List[str]:
    """
    Parses a comma separated string into a list

    Argumments:
        value (str or List[Any]):
            the value as a string or list of strings

    Returns:
        list[str]:
            the parsed value
    """
    if isinstance(value, list):
        return value

    parse_exception = EnvironmentVariableParseException(
        "Expected value in {name}={value} to be a list literal".format(
            name=name, value=value
        )
    )

    try:
        parsed_value = literal_eval(value)
    except (ValueError, SyntaxError) as ex:
        raise parse_exception from ex

    if not isinstance(parsed_value, list):
        raise parse_exception

    return parsed_value


def parse_delimited_list(name, value, default):
    """
    Parses a comma separated string into a list

    Argumments:
        value (str or list[str]):
            the value as a string or list of strings

    Returns:
        list[str]:
            the parsed value
    """
    parsed_value = value
    if isinstance(value, str):
        parsed_value = value.split(",")

    return [item.strip(" ") for item in parsed_value]


class EnvParser:
    """Stateful tracker for environment variable parsing"""

    def __init__(self):
        self.reload()

    def reload(self):
        """Reloads the environment"""
        self._env = dict(os.environ)
        self._configured_vars = {}

    def validate(self):
        """
        Validates the current configuration

        Raises:
            ImproperlyConfigured:
                If any settings are missing
        """
        missing_settings = []

        for env_var in self._configured_vars.values():
            if env_var.required and env_var.value in (None, ""):
                missing_settings.append(env_var.name)

        if missing_settings:
            raise ImproperlyConfigured(
                "The following settings are missing: {}. You need to add these environment variables in .env file.".format(
                    ", ".join(missing_settings)
                )
            )

    def list_environment_vars(self) -> List[EnvVariable]:
        """
        Get the list of EnvVariables

        Returns:
            list of EnvVariable:
                the list of available env vars
        """
        return self._configured_vars.values()

    def get_features(self, prefix: str = "FEATURE_") -> Dict[str, bool]:
        """
        Get the list of features enabled for this app

        Args:
            prefix (str):
                feature prefix string

        Returns:
            dict of bool:
                dictionary of feature flags
        """
        return {
            key[len(prefix) :]: self.get_bool(
                name=key,
                default=False,
                description=f"Feature {key}",
                write_app_json=False,
            )
            for key in list(self._env.keys())
            if key.startswith(prefix)
        }

    def init_app_settings(self, *, gbs: Dict, namespace: str, site_name: str):
        """
        Configure the app static settings

        Args:
            gbs (dict):
                A dictionary of global variables where we can set settings
            namespace (str):
                the app settings namespace
            site_name (str):
                The name of the site
        """
        gbs["APP_SETTINGS_NAMESPACE"] = self.get_string(
            name="APP_SETTINGS_NAMESPACE",
            default=namespace,
            description="App environment variable namespace",
            write_app_json=False,
        )
        gbs["SITE_NAME"] = self.get_string(
            name="SITE_NAME", default=site_name, description="Site name"
        )

    def get_site_name(self):
        """Return the site name"""
        site_name = self._configured_vars.get("SITE_NAME")
        if not site_name:
            raise ImproperlyConfigured(
                "Site name isn't set, add a call to init_app_settings()"
            )
        return site_name.value

    def app_namespaced(self, setting_key: str) -> str:
        """
        Return a namespaced setting key

        Example: app_namespaced("KEY") -> "APP_KEY"
        """
        namespace = self._configured_vars.get("APP_SETTINGS_NAMESPACE")
        if not namespace:
            raise ImproperlyConfigured(
                "App settings namespace isn't set, add a call to init_app_settings()"
            )
        return f"{namespace.value}_{setting_key}"

    get_string = var_parser(parse_str)
    get_bool = var_parser(parse_bool)
    get_int = var_parser(parse_int)
    get_list_literal = var_parser(parse_list_literal)
    get_delimited_list = var_parser(parse_delimited_list)


env = EnvParser()

# methods below are our exported module interface
get_string = env.get_string
get_int = env.get_int
get_bool = env.get_bool
get_list_literal = env.get_list_literal
get_delimited_list = env.get_delimited_list
reload = env.reload
validate = env.validate
list_environment_vars = env.list_environment_vars
get_features = env.get_features
init_app_settings = env.init_app_settings
app_namespaced = env.app_namespaced
get_site_name = env.get_site_name


def import_settings_modules(gbs: Dict, *module_names: str):
    """
    Import settings from modules

    Args:
        gbs (dict):
            the value of a calling module's globals()

    Usage:
        import_settings_modules(globals(), "module1.settings", "module2.settings")
    """

    # this function imports modules and then walks each, adding uppercased vars
    # to the calling settings module (typically settings.py in a django project)
    for module_name in module_names:
        mod = importlib.import_module(module_name)
        for setting_name in dir(mod):
            if setting_name.isupper():
                setting_value = getattr(mod, setting_name)
                gbs[setting_name] = setting_value


def generate_app_json():
    """
    Generate a new app.json data structure in-memory using app.base.json and settings.py

    Returns:
        dict:
            object that can be serialized to JSON for app.json
    """
    with open("app.base.json") as app_template_json:
        config = json.load(app_template_json)

    for env_var in list_environment_vars():
        if env_var.dev_only or not env_var.write_app_json:
            continue

        if env_var.name not in config["env"]:
            config["env"][env_var.name] = {}

        config["env"][env_var.name].update(
            {"description": env_var.description, "required": env_var.required}
        )

    return config



"""Webpack utilities"""
from django.conf import settings
from django.http import HttpRequest
from django.templatetags.static import static

from mitol.common.utils.urls import ensure_trailing_slash


def webpack_public_path(request: HttpRequest) -> str:
    """
    Return the correct public_path for Webpack to use
    """
    if settings.WEBPACK_USE_DEV_SERVER:
        return ensure_trailing_slash(webpack_dev_server_url(request))
    else:
        return ensure_trailing_slash(static("bundles/"))


def webpack_dev_server_host(request: HttpRequest) -> str:
    """
    Get the correct webpack dev server host
    """
    return settings.WEBPACK_DEV_SERVER_HOST or request.get_host().split(":")[0]


def webpack_dev_server_url(request: HttpRequest) -> str:
    """
    Get the full URL where the webpack dev server should be running
    """
    return "http://{}:{}".format(
        webpack_dev_server_host(request), settings.WEBPACK_DEV_SERVER_PORT
    )



"""Boilerplate webpack settings parsing"""
# pragma: no cover
from mitol.common.envs import get_bool, get_int, get_string

WEBPACK_DEV_SERVER_HOST = get_string(
    name="WEBPACK_DEV_SERVER_HOST",
    default="",
    dev_only=True,
    description="The webpack dev server hostname, development only",
)
WEBPACK_DEV_SERVER_PORT = get_int(
    name="WEBPACK_DEV_SERVER_PORT",
    default=None,
    dev_only=True,
    description="The webpack dev server port, development only",
)
WEBPACK_DISABLE_LOADER_STATS = get_bool(
    name="WEBPACK_DISABLE_LOADER_STATS",
    default=False,
    dev_only=True,
    description="Disables the webpack loader, development environment only.",
)
WEBPACK_USE_DEV_SERVER = get_bool(
    name="WEBPACK_USE_DEV_SERVER",
    default=False,
    dev_only=True,
    description="Enables the webpack devserver, development only",
)



"""Requests utilities"""
import logging
from http import HTTPStatus

import requests
from requests import Response

log = logging.getLogger(__name__)


def get_error_response_summary(response: Response) -> str:
    """
    Returns a summary of an error raised from a failed HTTP request using the requests library

    Args:
        response (requests.Response): The requests library response object

    Returns:
        str: A summary of the error response
    """
    # If the response is an HTML document, include the URL in the summary but not the raw HTML
    if "text/html" in response.headers.get("Content-Type", ""):
        summary_dict = {"url": response.url, "content": "(HTML body ignored)"}
    else:
        summary_dict = {"content": response.text}
    summary_dict_str = ", ".join([f"{k}: {v}" for k, v in summary_dict.items()])
    return f"Response - code: {response.status_code}, {summary_dict_str}"


def is_json_response(response: Response) -> bool:
    """
    Returns True if the given response object is JSON-parseable

    Args:
        response (requests.Response): The requests library response object

    Returns:
        bool: True if this response is JSON-parseable
    """
    return response.headers.get("Content-Type") == "application/json"


def request_get_with_timeout_retry(url: str, retries: int) -> Response:
    """
    Makes a GET request, and retries if the server responds with a 504 (timeout)

    Args:
        url (str): The URL of the Mailgun API endpoint
        retries (int): The number of times to retry the request

    Returns:
        response (requests.models.Response): The requests library response object

    Raises:
        requests.exceptions.HTTPError: Raised if the response has a status code indicating an error
    """
    resp = requests.get(url)
    # If there was a timeout (504), retry before giving up
    tries = 1
    while resp.status_code == HTTPStatus.GATEWAY_TIMEOUT and tries <= retries:
        tries += 1
        log.warning(
            "GET request timed out (%s). Retrying for attempt %d...", url, tries
        )
        resp = requests.get(url)
    resp.raise_for_status()
    return resp



""" utils """
# Backward compatability
from mitol.common.utils.collections import *  # noqa: F401,F403
from mitol.common.utils.currency import *  # noqa: F401,F403
from mitol.common.utils.datetime import *  # noqa: F401,F403
from mitol.common.utils.http_requests import *  # noqa: F401,F403
from mitol.common.utils.urls import *  # noqa: F401,F403
from mitol.common.utils.webpack import *  # noqa: F401,F403



"""URL utilities"""
from urllib.parse import ParseResult, urlparse, urlunparse


def ensure_trailing_slash(url: str) -> str:
    """ensure a url has a trailing slash"""
    return url if url.endswith("/") else url + "/"


def remove_password_from_url(url: str) -> str:
    """
    Remove a password from a URL

    Args:
        url (str): A URL

    Returns:
        str: A URL without a password
    """
    pieces = urlparse(url)
    netloc = pieces.netloc
    userinfo, delimiter, hostinfo = netloc.rpartition("@")
    if delimiter:
        username, _, _ = userinfo.partition(":")
        rejoined_netloc = f"{username}{delimiter}{hostinfo}"
    else:
        rejoined_netloc = netloc

    return urlunparse(
        ParseResult(
            scheme=pieces.scheme,
            netloc=rejoined_netloc,
            path=pieces.path,
            params=pieces.params,
            query=pieces.query,
            fragment=pieces.fragment,
        )
    )




# DO NOT EDIT THIS FILE -- AUTOGENERATED BY PANTS
# Target: src/mitol/common:mitol-django-common

from setuptools import setup

setup(**{
    'authors': [
        'MIT Office of Open Learning <mitx-devops@mit.edu>',
    ],
    'classifiers': [
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    'description': 'MIT Open Learning django app extensions for common utilities',
    'install_requires': (
        'django-webpack-loader>=0.7.0',
        'django<4.0,>=2.2.12',
        'factory-boy~=3.2',
        'pytest==6.1.2',
        'pytz>=2020.4',
        'requests>=2.20.0',
        'setuptools',
    ),
    'license': 'BSD 3-Clause License',
    'name': 'mitol-django-common',
    'namespace_packages': (
        'mitol',
    ),
    'package_data': {
        'mitol.common': (
            'CHANGELOG.md',
            'README.md',
            'py.typed',
        ),
    },
    'packages': (
        'mitol',
        'mitol.common',
        'mitol.common.factories',
        'mitol.common.management',
        'mitol.common.management.commands',
        'mitol.common.settings',
        'mitol.common.templatetags',
        'mitol.common.utils',
    ),
    'version': '2.2.4',
    'zip_safe': True,
})



"""Datetime utilities"""
from datetime import datetime, timedelta

import pytz


def now_in_utc() -> datetime:
    """
    Get the current time in UTC
    Returns:
        datetime.datetime: A datetime object for the current time
    """
    return datetime.now(tz=pytz.UTC)


def is_near_now(time: datetime, epsilon: timedelta = timedelta(seconds=5)) -> bool:
    """
    Returns true if time is within five seconds or so of now
    Args:
        time (datetime.datetime):
            The time to test
        epsilon (datetime.timedelta):
            The allowed delta in time
    Returns:
        bool:
            True if near now, false otherwise
    """
    now = now_in_utc()
    return now - epsilon < time < now + epsilon
